@@ -24,6 +24,8 @@
 import dbus
 import os as _os
 import Queue
+import threading
+import time
 import traceback
 
 from busutils import *
@@ -133,10 +135,12 @@ class Registry(object):
 
                 factory = AccessibleFactory(cache)
 
+                self.has_implementations = True
+
                 _os.environ["AT_SPI_CLIENT"] = "1"
 
                 # Set up the device event controllers
-                _connection = SyncAccessibilityBus ()
+                _connection = SyncAccessibilityBus (self)
                 _bus_object = _connection.get_object("org.freedesktop.DBus", "/org/freedesktop/DBus")
 
                 if app_name:
@@ -153,7 +157,7 @@ class Registry(object):
                         self.desktop = factory (name, _ATSPI_ROOT_PATH, _ATSPI_DESKTOP)
 
 		self.async = False	# not fully supported yet
-                self.has_implementations = True
+                self.started = False
 
         def _set_default_registry (self):
                 self._set_registry (MAIN_LOOP_GLIB)
@@ -173,10 +177,40 @@ class Registry(object):
                 """
                 if not self.has_implementations:
                         self._set_default_registry ()
-                try:
+                self.acquireLock()
+                self.thread = threading.currentThread()
+                self.started = True
+
+                def idleReleaseLock():
+                        self.releaseLock()
+                        return False
+
+                gobject.idle_add(idleReleaseLock)
+
+                if gil:
+                        def releaseGIL():
+                                try:
+                                        time.sleep(1e-5)
+                                except KeyboardInterrupt, e:
+                                        # store the exception for later
+                                        releaseGIL.keyboard_exception = e
+                                        self.stop()
+                                return True
+                        # make room for an exception if one occurs during the 
+                        releaseGIL.keyboard_exception = None
+                        i = gobject.idle_add(releaseGIL)
+                        
                         self.main_loop.run()
-                except KeyboardInterrupt:
-                        pass
+                        self.started = False
+                        gobject.source_remove(i)
+                        if releaseGIL.keyboard_exception is not None:
+                                # raise an keyboard exception we may have gotten earlier
+                                raise releaseGIL.keyboard_exception
+                else:
+                        try:
+                                self.main_loop.run()
+                        except KeyboardInterrupt:
+                                pass
 
         def stop(self, *args):
                 """
@@ -390,6 +424,22 @@ class Registry(object):
                         self._set_default_registry ()
                 self.device_event_register.generateMouseEvent (x, y, name)
 
+        def acquireLock(self):
+                """
+                Acquire a lock, creating the registry irst if needed.
+                """
+                try:
+                        self.lock.acquire()
+                except AttributeError:
+                        self.lock = threading.RLock()
+                        self.lock.acquire()
+
+        def releaseLock(self):
+                """
+                Release the lock
+                """
+                self.lock.release()
+
 #------------------------------------------------------------------------------
 
 def set_default_registry (main_loop, app_name=None):
